﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Problems.VehicleRouting.Encodings.Potvin;
using HeuristicLab.Problems.VehicleRouting.Encodings;
using HeuristicLab.Problems.VehicleRouting;
using HeuristicLab.Problems.VehicleRouting.Interfaces;

namespace HeuristicLab.Analysis.FitnessLandscape.VRP {
  [Item("TwoOptStarManipulator", "Two opt star manipulation")]
  [StorableClass]
  public sealed class TwoOptStarManipulator : PotvinManipulator {
    [StorableConstructor]
    private TwoOptStarManipulator(bool deserializing) : base(deserializing) { }
    private TwoOptStarManipulator(TwoOptStarManipulator original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new TwoOptStarManipulator(this, cloner);
    }
    public TwoOptStarManipulator()
      : base() {
    }

    public static void Apply(IRandom random, PotvinEncoding individual, IVRPProblemInstance problemInstance, bool allowInfeasible) {
      bool feasible;

      //consider creating new tour
      individual.Tours.Add(new Tour());

      do {
        feasible = true;

        int route1Idx = random.Next(individual.Tours.Count);
        int route2Idx = random.Next(individual.Tours.Count - 1);
        if (route2Idx >= route1Idx)
          route2Idx++;

        Tour route1 = individual.Tours[route1Idx];
        Tour route2 = individual.Tours[route2Idx];

        int x1 = random.Next(route1.Stops.Count + 1);
        int x2 = random.Next(route2.Stops.Count + 1);

        if (!allowInfeasible) {
          bool originalFeasible = problemInstance.TourFeasible(route1, individual) &&
                                  problemInstance.TourFeasible(route2, individual);

          if (originalFeasible) {
            int count = route1.Stops.Count - x1;
            List<int> segmentX1 = new List<int>();
            if (count > 0) {
              segmentX1 = route1.Stops.GetRange(x1, count);
              route1.Stops.RemoveRange(x1, count);
            }

            count = route2.Stops.Count - x2;
            List<int> segmentX2 = new List<int>();
            if (count > 0) {
              segmentX2 = route2.Stops.GetRange(x2, count);
              route2.Stops.RemoveRange(x2, count);
            }

            route1.Stops.AddRange(segmentX2);
            feasible = problemInstance.TourFeasible(route1, individual);
            if (feasible) {
              route2.Stops.AddRange(segmentX1);
              feasible = problemInstance.TourFeasible(route2, individual);
              if (!feasible) {
                // undo all changes
                route1.Stops.RemoveRange(x1, segmentX2.Count);
                route1.Stops.AddRange(segmentX1);
                route2.Stops.RemoveRange(x2, segmentX1.Count);
                route2.Stops.AddRange(segmentX2);
              }
            } else {
              // undo changes 
              route1.Stops.RemoveRange(x1, segmentX2.Count);
              route1.Stops.AddRange(segmentX1);
              route2.Stops.AddRange(segmentX2);
            }

          } else {
            feasible = false;
          }
        }
      } while (!feasible);

      individual.Tours.RemoveAll(t => t.Stops.Count == 0);
    }

    protected override void Manipulate(IRandom random, PotvinEncoding individual) {
      bool allowInfeasible = AllowInfeasibleSolutions.Value.Value;
      Apply(random, individual, ProblemInstance, allowInfeasible);
    }
  }
}
